import os
import xml.etree.ElementTree as ET


class SimConfig:
    """
    This class will prepare a configuration file for SUMO simulations.

    Attributes
    ----------
    config_path: str
        path to xml configuration file

    Methods
    -------
    change_netfile(self, netfile_name):
        Changes the network file name in configuration file

    save(self, fname):
        Writes the xml datastructure into a file on the disk 

    """

    def __init__(self, config_path):
        self.config_path = config_path
        self._load_xml()

    def _load_xml(self):
        self.tree = ET.parse(self.config_path)
        self.root = self.tree.getroot()

    def change_netfile(self, netfile_name):
        self.root[0][0].attrib['value'] = netfile_name 
    
    def save(self, fname):
        self.tree.write(fname)



class SimulationHandler:
    """
    Run a SUMO simulation to visualize the city taffic flow and generate raw outputs.

    Attributes
    ----------


    Methods
    -------
    run(self, gui):
        opens a process to run simualtion

    """

    def __init__(self, config_path):
        self.config_path = config_path

    def run(self, output_tripinfo='tripinfo.out', gui=False):
        if gui == False:
            os.popen(f"sumo -c {self.config_path} --tripinfo-output {output_tripinfo}").read()
        else:
            os.popen(f"sumo-gui -c {self.config_path}").read()
        print("Simulation finished.")



from sumo_pts.map.city import City


aachen = City('osm.net.xml', 'osm_pt.rou.xml')

aachen.change_street_speed('Pontwall', 0.1)
aachen.save('osm_slow.net.xml')

street_names = ['Pontwall', 'Seffenter Weg', 'Forckenbeckstraße', 'Halifaxstraße',
               'Alexianergraben', 'Franzstraße', 'Jakobstraße', 'Vaalser Straße',
               'Komphausbadstraße', 'Peterstraße', 'Heinrichsallee', 'Theaterstraße',
               'Saarstraße', 'Boxgraben', 'Roermonder Straße', 'Hirschgraben']

aachen.show_statistics()



from sumo_pts.output.tripinfo_reader import TripInfo


tripinfo_output = TripInfo('tripinfo.out', 50)
total_delay = tripinfo_output.get_total_delay()
num_delays = tripinfo_output.get_num_delay()

print(f'{num_delays} delays happend. in total {total_delay:.2f} seconds. average {total_delay/num_delays:.2f} seconds')



from sumo_pts.map.city import City
from sumo_pts.simulation import SimulationHandler, SimConfig
from sumo_pts.output.tripinfo_reader import TripInfo

from scipy.stats import qmc
import numpy as np



stree_names = ['Pontwall', 'Seffenter Weg', 'Forckenbeckstraße', 'Halifaxstraße',
               'Alexianergraben', 'Franzstraße', 'Jakobstraße', 'Vaalser Straße',
               'Komphausbadstraße', 'Peterstraße', 'Heinrichsallee', 'Theaterstraße',
               'Saarstraße', 'Boxgraben', 'Roermonder Straße', 'Hirschgraben']

sampler = qmc.LatinHypercube(d=len(stree_names))
sample = sampler.random(n=10)


for i in range(sample.shape[0]):
    aachen = City('osm.net.xml', 'osm_pt.rou.xml')
    for j, street in enumerate(stree_names):
        aachen.change_street_speed(street, sample[i, j])
    aachen.save(f'./data/osm_slow_{i}.net.xml')


for i in range(sample.shape[0]):
    conf = SimConfig('aachen.sumocfg')
    conf.change_netfile(f'./data/osm_slow_{i}.net.xml')
    conf.save('aachen_1.sumocfg')

    acchen_sim = SimulationHandler("aachen_1.sumocfg")
    acchen_sim.run(output_tripinfo=f'./output/tripinfo_{i}.out', gui=False)


total_delays = []
for i in range(sample.shape[0]):
    tripinfo_output = TripInfo(f'./output/tripinfo_{i}.out', 0)
    total_delay = tripinfo_output.get_total_delay()
    num_delays = tripinfo_output.get_num_delay()
    print(f'{num_delays} delays happend. in total {total_delay:.2f} seconds. average {total_delay/num_delays:.2f} seconds')
    total_delays.append(total_delay)


total_delays = np.array(total_delays)

np.save('input.npy', sample)
np.save('output.npy', total_delays)



import xml.etree.ElementTree as ET


class TripInfo:
    """
    Read trip information output of SUMO simualtion.

    Attributes
    ----------
    file_path: str
        path to ouput file
    delay_threshold: float or int = 0
        time threshold (seconds) to consider delay

    Methods
    -------
    get_root(self):
        returns root node of xml file
    get_total_delay(self):
        returns sum of all delays above a given threshold

    """

    def __init__(self, file_path, delay_threshold=0):
        self.file_path = file_path
        self.delay_threshold = delay_threshold
        self._load_xml()

    def _load_xml(self):
        tree = ET.parse(self.file_path)
        self.root = tree.getroot()

    def get_root(self):
        return self.root

    def get_total_delay(self):
        self.total_delay = 0 
        self.count_delays = 0 
        for child in self.root:
            delay = float(child.attrib['departDelay'])
            if delay > self.delay_threshold:
                self.total_delay += delay
                self.count_delays += 1
                # print(child.attrib['id'], child.attrib['departDelay'])
        # print(self.total_delay, self.count_delays)
        return self.total_delay
    
    def get_num_delay(self):
        return self.count_delays

# tree = ET.parse('tripinfo.out')
# root = tree.getroot()

# for c in root:
#     if float(c.attrib['departDelay']) > 50: 
#         print(c.attrib['id'], c.attrib['departDelay'])


import time

import numpy as np
import tensorflow as tf
from sklearn.neural_network import MLPRegressor
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error, r2_score
import matplotlib.pyplot as plt
from ray import train, tune
from ray.tune.schedulers import PopulationBasedTraining


class MyCallback(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs={}):
        train.report({"val_loss": logs.get('val_loss')})


def train_model(config):
    number_epochs = 1000
    batch_size = 160

    number_layers = config['n_l']
    lr = config['lr']
    
    model = tf.keras.Sequential()
    model.add(tf.keras.layers.Dense(X_train.shape[1]))
    
    for i in range(number_layers):
        model.add(tf.keras.layers.Dense(config[f"hidden{i+1}"], activation='sigmoid'))
    
    model.add(tf.keras.layers.Dense(y_train.shape[1]))
    
    optt = tf.keras.optimizers.Adam(learning_rate=lr)  

    model.compile(loss='mse',
                    optimizer=optt,
                    metrics=['mae'])
    
    history = model.fit(X_train, y_train_s,
                    epochs=number_epochs, batch_size=batch_size, verbose=True,
                    validation_data=(X_test, y_test_s),
                    callbacks=[MyCallback()])

    val_loss = history.history["val_loss"][-1]
    return {"val_loss": val_loss, 'model': model, 'history': history}


def tune_model():
    
    pbt = PopulationBasedTraining(
    time_attr="training_iteration",
    perturbation_interval=200,  
    hyperparam_mutations={
            "n_l": tune.choice([2, 3]),
            "lr": tune.loguniform(1e-5, 1e-1),
            "hidden1": tune.uniform(10, 100),
            "hidden2": tune.uniform(10, 100),
            "hidden3": tune.uniform(10, 100),
        },
    )

    tuner = tune.Tuner(
        train_model,
        tune_config=tune.TuneConfig(
            metric="val_loss",
            mode="min",
            scheduler=pbt,
            num_samples=50,
        ),
        run_config=train.RunConfig(
            name="my_exp1",
        )
    )
    
    t_i = time.time()
    results = tuner.fit()
    HP_fit_time = time.time() - t_i
    print('elapsed time in min: ', HP_fit_time/60)

    print("Best hyperparameters found were: ", results.get_best_result().config)
    return results


data_X = np.load('input.npy')
data_y = np.load('output.npy').reshape(-1, 1)

X_train, X_test, y_train, y_test = train_test_split(data_X, data_y, test_size=0.2, 
                                                    random_state=37)

scaler = MinMaxScaler()
y_train_s = scaler.fit_transform(y_train)
y_test_s = scaler.transform(y_test)


hp_results = tune_model()
best_nn_hp = hp_results.get_best_result().config
# best_nn_hp = {'n_l': 3, 'lr': 0.008038345356318897, 'hidden1': 87.15517437945864, 'hidden2': 59.70976351699706, 'hidden3': 29.935028055960178}
# best_nn_hp = {'n_l': 3, 'lr': 0.0007038345356318897, 'hidden1': 87.15517437945864, 'hidden2': 59.70976351699706, 'hidden3': 29.935028055960178}

train_res = train_model(best_nn_hp)
model = train_res["model"]
nn_model_hist = train_res["history"]
print("\nBest hyperparameters found were: ", best_nn_hp)

plt.figure()
plt.plot(nn_model_hist.history['loss'])
plt.plot(nn_model_hist.history['val_loss'])
plt.legend(['train set', 'test set'])
plt.xlabel('# epoch')
plt.ylabel('RMS Error')
plt.yscale('log')
plt.title('Leaning curve')
plt.grid(True)
plt.savefig('learningC_HP.png', dpi=550)
plt.show()

y_train_predicted = model.predict(X_train)
y_test_predicted = model.predict(X_test)

print(f'RMSE: train= {mean_squared_error(y_train_s, y_train_predicted)}, test= {mean_squared_error(y_test_s, y_test_predicted)}')
print(f'R2 score: train= {r2_score(y_train_s, y_train_predicted)}, test= {r2_score(y_test_s, y_test_predicted)}')




import xml.etree.ElementTree as ET


class City:
    """
    Model roads and bus transportaion system of a city.

    Attributes
    ----------
    net_file_path: str
        path to xml network file

    route_file_path: str
        path to xml route file

    Methods
    -------
    load_xml(self, path):
        returns the thermal diffusivity of the material

    """
    
    def __init__(self, net_file_path, route_file_path):
        self.net_file_path = net_file_path
        self.route_file_path = route_file_path
        self._load_xml()

    def _load_xml(self):
        self.net_tree = ET.parse(self.net_file_path)
        self.net_root = self.net_tree.getroot()
        self.route_tree = ET.parse(self.route_file_path)
        self.route_root = self.route_tree.getroot()

    def get_net_root(self):
        return self.net_root

    def get_route_root(self):
        return self.route_root

    def change_street_speed(self, street_name, speed_scale):
        for child1 in self.net_root:
            if child1.tag == 'edge':
                if 'name' in child1.attrib.keys(): 
                    if child1.attrib['name'] == street_name:
                        for child2 in child1:
                            if 'speed' in child2.attrib.keys():
                                # print(child2.attrib['speed'])
                                current_speed = float(child2.attrib['speed'])
                                child2.attrib['speed'] = str(speed_scale * current_speed)
                                # print(child2.attrib['speed'])
    
    def save(self, fname):
        self.net_tree.write(fname)

    def show_statistics(self, street_names=None):
        """
        Printing information about the bus trips through city streets. 

        Parameters
        ----------
        street_names = None : list of str
            a list containing the name of the streets to report the trips.
        
        Returns
        -------
        counter_edge_bus_sorted: dict
            a dictionay mapping street names (str) to number of bus trips (int).
        """
        edge_id_name = {}
        for child1 in self.net_root:
            if child1.tag == 'edge':
                if 'name' in child1.attrib.keys():
                    # if child1.attrib['name'] == edge_id_name:
                    edge_id_name[child1.attrib['id']] = child1.attrib['name']

        # return edge_id_name
        counter_edge_bus = {}
        for child1 in self.route_root:
            if child1.tag == 'route':
                if 'edges' in child1.attrib.keys():
                    for edge_id in child1.attrib['edges'].split():
                        if edge_id in edge_id_name.keys():
                            street_name = edge_id_name[edge_id]
                            if street_name in counter_edge_bus.keys():
                                counter_edge_bus[street_name] += 1
                            else:
                                counter_edge_bus[street_name] = 1
        
        counter_edge_bus_sorted = {k: v for k, v in sorted(counter_edge_bus.items(), key=lambda item: -item[1])}
        total_trips = 0
        
        print('street name \t\t # bus trips')
        print('----------- \t\t   ------------')
        for street_name in counter_edge_bus_sorted:
            if street_names != None:
                if street_name in street_names:
                    print(f'{street_name} \t\t\t {counter_edge_bus_sorted[street_name]}')
            else:
                print(f'{street_name} \t\t\t {counter_edge_bus_sorted[street_name]}')
            
            total_trips += counter_edge_bus_sorted[street_name]
        print(f'total # trips: \t\t\t {total_trips}')
        
        return counter_edge_bus_sorted



import numpy as np
from sklearn.neural_network import MLPRegressor
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error, r2_score
import matplotlib.pyplot as plt


data_X = np.load('input.npy')
data_y = np.load('output.npy').reshape(-1, 1)

X_train, X_test, y_train, y_test = train_test_split(data_X, data_y, test_size=0.2, 
                                                    random_state=37)

scaler = MinMaxScaler()
y_train_s = scaler.fit_transform(y_train)
y_test_s = scaler.transform(y_test)


model = MLPRegressor(hidden_layer_sizes=[20, 20],
                     random_state=1,
                     activation='relu',
                     learning_rate_init=0.1,
                     solver='sgd',
                     tol=1e-10,
                     alpha=0.1,
                     max_iter=1000).fit(X_train, y_train_s.reshape(-1))

y_train_predicted = model.predict(X_train)
y_test_predicted = model.predict(X_test)

print(f'RMSE: train= {mean_squared_error(y_train_s, y_train_predicted)}, test= {mean_squared_error(y_test_s, y_test_predicted)}')
print(f'R2 score: train= {r2_score(y_train_s, y_train_predicted)}, test= {r2_score(y_test_s, y_test_predicted)}')

plt.plot(model.loss_curve_)
plt.xlabel('# epoch')
plt.ylabel('RMSE(train set)')
plt.yscale('log')
plt.title('Leaning curve')
plt.grid(True)
plt.savefig('learningC.png', dpi=550)
plt.show()



from sumo_pts.simulation import SimulationHandler, SimConfig


# conf = SimConfig('aachen.sumocfg')
# conf.change_netfile('osm.net.xml.gz')
# conf.save('aachen.sumocfg.deleteme')

acchen_sim = SimulationHandler("aachen.sumocfg")
acchen_sim.run(gui=True)



from setuptools import setup


with open('README.md', 'r') as f:
      description = f.read()

setup(name='sumo_pts',
      version='0.1',
      description='A Python interface for simulating public transport systems using SUMO.',
      url='https://github.com/itsAliSali/sumo_pts',
      author='itsAliSali',
      author_email='ali.salimian@rwth-aachen.de',
      license='EPL2',
      packages=['sumo_pts'],
      zip_safe=False,
      long_description=description,
      long_description_content_type='text/markdown',
)



import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error, r2_score
import matplotlib.pyplot as plt

from sumo_pts.map.city import City


aachen = City('osm.net.xml', 'osm_pt.rou.xml')

street_names = ['Pontwall', 'Seffenter Weg', 'Forckenbeckstraße', 'Halifaxstraße',
               'Alexianergraben', 'Franzstraße', 'Jakobstraße', 'Vaalser Straße',
               'Komphausbadstraße', 'Peterstraße', 'Heinrichsallee', 'Theaterstraße',
               'Saarstraße', 'Boxgraben', 'Roermonder Straße', 'Hirschgraben']

bus_freq = aachen.show_statistics(street_names)

data_X = np.load('input.npy')
data_y = np.load('output.npy').reshape(-1, 1)

X_train, X_test, y_train, y_test = train_test_split(data_X, data_y, test_size=0.2, 
                                                    random_state=37)

scaler = MinMaxScaler()
y_train_s = scaler.fit_transform(y_train)
y_test_s = scaler.transform(y_test)


model = LinearRegression()
model.fit(X_train, y_train_s)
importance = np.abs(model.coef_[0])
for i,v in enumerate(importance):
    print('Feature: %0d, Score: %.5f' % (i,v))

bar = plt.bar([x for x in range(len(importance))], importance)

for i, rect in enumerate(bar):
    height = rect.get_height()
    plt.text(rect.get_x() + rect.get_width() / 2.0, height, str(bus_freq[street_names[i]]), ha='center', va='bottom')

plt.xticks([x for x in range(len(importance))], street_names, rotation=90)
plt.ylabel('Regression coefficient')
plt.show()

from sklearn.ensemble import RandomForestRegressor

feature_names = [f"feature {i}" for i in range(X_train.shape[1])]
forest = RandomForestRegressor(random_state=0)
forest.fit(X_train, y_train_s.ravel())

importances = forest.feature_importances_
std = np.std([tree.feature_importances_ for tree in forest.estimators_], axis=0)

import pandas as pd

forest_importances = pd.Series(importances, index=feature_names)

fig, ax = plt.subplots()
forest_importances.plot.bar(yerr=std, ax=ax)
ax.set_title("Feature importances using MDI")
ax.set_ylabel("Mean decrease in impurity")
fig.tight_layout()